﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

public class Field : MonoBehaviour
{
    // Start is called before the first frame update
    public GameObject block;             // フィールドに敷き詰めるブロックの元
    public List<GameObject> list_block;    // 敷き詰めたブロックの配列
    public BlockSet blockSet;            // ブロックセットの元

    public GameObject scoreObject;  // スコア表示のテキスト
    public GameObject bombObject;   // ボム数表示のテキスト

    // 効果音ファイル(IDEで関連付け)
    public AudioClip sndRotate, sndPut, sndErase, sndBomb;

    // フィールドのサイズ
    const int FIELD_X = 5;
    const int FIELD_Y = 10;
    const int FIELD_Z = 5;

    // キーを押しているフラグ(連打防止用)
    bool keyW;
    bool keyA;
    bool keyS;
    bool keyD;
    bool keyE;
    bool keyQ;
    bool keyB;
    bool keySpace;

    // 落下中のブロックセットの位置
    Vector3 bpos;

    // ブロックセットのインスタンス
    BlockSet bs;

    // スコア、ボム
    int score;
    int bomb;

    // 効果音オブジェクト
    AudioSource audioSource;

    // 回転位置を取得するためのカメラオブジェクト(視点によりキー操作を変える)
    Camera cam;
    GameObject camobj;

    void Start()
    {
        // Debug.Log("Start!");

        // 初期化処理
        bpos.x = 1;         // アクティブブロックの初期位置X
        bpos.y = FIELD_Y;
        bpos.z = 1;

        score = 0;          // スコア
        bomb = 3;           // ボム数

        block.SetActive(false);

        // 効果音を扱うオブジェクトを取得(IDEで追加しておく)
        audioSource = GetComponent<AudioSource>();

        // 回転角度を得るためにカメラを取得
        camobj = GameObject.Find("Main Camera");
        cam = camobj.GetComponent<Camera>();


        // フィールドにブロックを敷き詰めて非表示にする
        // フィールドブロックを動的に生成
        int pos, i, x, y, z;
        i = 0;
        for (y = 0; y < FIELD_Y; y++)
        {
            for (z = 0; z < FIELD_Z; z++)
            {
                for (x = 0; x < FIELD_X; x++)
                {
                    pos = PosToOrder(x, y, z);
                    if(pos == -1)
                    {
                        Debug.Log("PosToOrder Error!");
                        return;
                    }
                    list_block.Add(Instantiate(block, new Vector3(x, y, z), transform.rotation));
                    list_block[i].SetActive(false);                    
                    i++;
                }

            }
        }

        // ブロックセットの生成(操作する落下中のブロック)
       bs =  Instantiate(blockSet, new Vector3(5, 10, 1), transform.rotation);
        
        // タイマーセット
        InvokeRepeating("TimeEvent", 1.0f, 2.0f);

        // Debug.Log("Start() OK!");
    }

    // タイマー処理
    void TimeEvent()
    {
        //Debug.Log("Timer Event!");

        // アクティブブロックの落下処理
        // bposはフィールド上のブロックセット起点の位置
        bpos.y -= 1f;

        if(IsMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
        {
            bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
            // bs.RotateY();
        }
        else
        {
            /* アクティブブロックが移動不可能だった場合、フィールドブロックとして固定する
               アクティブブロックのリストを取得 → フィールド座標に変換 → 座標のデータ位置を取得 → 固定 */
            
            bpos.y += 1;    // 座標を戻す

            // フィールド位置(オフセット)を与えて、実際のブロックの座標を得る(構成されるブロックの座標リスト)
            List<Vector3> list = new List<Vector3>();
            list = bs.GetActiveBlockArray((int)bpos.x, (int)bpos.y, (int)bpos.z);

            for (int i = 0; i < list.Count; i++)
            {
                // 座標からデータ位置に変換(フィールド配列の何番目のブロックか）
                int order = PosToOrder((int)list[i].x, (int)list[i].y, (int)list[i].z);
                if (order != -1)
                {
                    // フィールドブロックとして固定する
                    list_block[order].SetActive(true);
                    list_block[order].GetComponent<Renderer>().material.color = bs.cr; // 色合わせ

                    // 効果音(設置音)
                    audioSource.PlayOneShot(sndPut);
                }
                else
                {
                    // 固定時に場外なのでゲームオーバー(演出は省略)
                    for (int j = 0; j < FIELD_Y; j++)
                    {
                        RemoveStep(0);
                    }
                    score = 0;
                    bomb = 3;
                }
            }

            // 揃い判定・消去処理(すべての段で揃っているか判定、揃っていたら消去)
            int cnt = 0;
            for(int y=0; y<FIELD_Y; y++)
            {
                if (IsFilledStep(y))
                {
                    RemoveStep(y);
                    y--;
                    cnt++;
                }
            }

            if (cnt > 0)
            {
                // 消去数に応じて得点を獲得(消去数の二乗×1000)
                score += cnt * cnt * 1000;

                // 消去数に応じてボムの回復
                bomb += cnt - 1;

                // 効果音(消去)
                audioSource.PlayOneShot(sndErase);
            }

            // 新しいブロックを生成(一番高い位置の中央）
            bpos.y = FIELD_Y;
            bpos.x = (FIELD_X - 3) / 2;
            bpos.z = FIELD_Z / 2;

            bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
            bs.Block_New();

        }
        /*
         float a = cam.transform.localRotation.eulerAngles.y;
         45~0 , 360~315: 正面
         315~225: 右視点
         225:135: 裏視点
         135:45: 左視点
         */
        UpdateUI();
    }

    void UpdateUI()
    {
        // テキストの更新
        Text score_text = scoreObject.GetComponent<Text>();
        Text bomb_text = bombObject.GetComponent<Text>();

        score_text.text = "score:" + score;
        bomb_text.text = "bomb:" + bomb;
    }

    // Update is called once per frame
    void Update()
    {

        Vector3 logpos;
        logpos = bpos;

        // ↑キー or Wキーを押したときの処理(ブロックを移動する)
        if (Input.GetKey(KeyCode.UpArrow) || Input.GetKey(KeyCode.W))
        {
            if (keyW == false)
            {
                // ブロックの移動（カメラ視点により移動方向を変える）
                float a = cam.transform.localRotation.eulerAngles.y;
                if (a <= 45 || a >= 315) // 正面
                {
                    bpos.z += 1;
                }
                else if (a <= 315 && a >= 225)    // 右視点
                {
                    bpos.x -= 1;
                }
                else if (a <= 225 && a >= 135)    // 裏視点
                {
                    bpos.z -= 1;
                }
                else if (a <= 135 && a >= 45)     // 左視点
                {
                    bpos.x += 1;
                }

                if (IsMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
                }
                else
                {
                    bpos = logpos;
                }
            }
            keyW = true;
        }
        else
        {
            keyW = false;
        }

        // ↓キー or Sキーを押したときの処理(ブロックを移動する)
        if (Input.GetKey(KeyCode.DownArrow) || Input.GetKey(KeyCode.S))
        {
            if (keyS == false)
            {
                float a = cam.transform.localRotation.eulerAngles.y;
                if (a <= 45 || a >= 315) // 正面
                {
                    bpos.z -= 1;
                }
                else if (a <= 315 && a >= 225)    // 右視点
                {
                    bpos.x += 1;
                }
                else if (a <= 225 && a >= 135)    // 裏視点
                {
                    bpos.z += 1;
                }
                else if (a <= 135 && a >= 45)     // 左視点
                {
                    bpos.x -= 1;
                }

                if (IsMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
                }
                else
                {
                    bpos = logpos;
                }
            }
            keyS = true;
        }
        else
        {
            keyS = false;
        }

        // ←キー or Aキーを押したときの処理(ブロックを移動する)
        if (Input.GetKey(KeyCode.LeftArrow) || Input.GetKey(KeyCode.A))
        {
            if (keyA == false)
            {
                float a = cam.transform.localRotation.eulerAngles.y;
                if (a <= 45 || a >= 315) // 正面
                {
                    bpos.x -= 1;
                }
                else if (a <= 315 && a >= 225)    // 右視点
                {
                    bpos.z -= 1;
                }
                else if (a <= 225 && a >= 135)    // 裏視点
                {
                    bpos.x += 1;
                }
                else if (a <= 135 && a >= 45)     // 左視点
                {
                    bpos.z += 1;
                }

                if (IsMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
                }
                else
                {
                    bpos = logpos;
                }
            }
            keyA = true;
        }
        else
        {
            keyA = false;
        }

        // →キー or Dキーを押したときの処理(ブロックを移動する)
        if (Input.GetKey(KeyCode.RightArrow) || Input.GetKey(KeyCode.D))
        {
            if (keyD == false)
            {
                float a = cam.transform.localRotation.eulerAngles.y;
                if (a <= 45 || a >= 315)          // 正面
                {
                    bpos.x += 1;
                }
                else if (a <= 315 && a >= 225)    // 右視点
                {
                    bpos.z += 1;
                }
                else if (a <= 225 && a >= 135)    // 裏視点
                {
                    bpos.x -= 1;
                }
                else if (a <= 135 && a >= 45)     // 左視点
                {
                    bpos.z -= 1;
                }

                if (IsMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.Block_Move((int)bpos.x, (int)bpos.y, (int)bpos.z);
                }
                else
                {
                    bpos = logpos;
                }
            }
            keyD = true;
        }
        else
        {
            keyD = false;
        }

        // Eキーでブロック回転処理
        if (Input.GetKey(KeyCode.E))
        {
            if (keyE == false)
            {
                bs.DummyRotateY();
                if (IsDummyMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.RotateY();
                    audioSource.PlayOneShot(sndRotate);
                }
            }
            keyE = true;
        }
        else
        {
            keyE = false;
        }

        // Qキーでブロック回転処理
        if (Input.GetKey(KeyCode.Q))
        {
            if (keyQ == false)
            {
                bs.DummyRotateX();
                if (IsDummyMove((int)bpos.x, (int)bpos.y, (int)bpos.z))
                {
                    bs.RotateX();
                    audioSource.PlayOneShot(sndRotate);
                }
            }
            keyQ = true;
        }
        else
        {
            keyQ = false;
        }

        // Bキーでボム使用(4段破壊)
        if (Input.GetKey(KeyCode.B))
        {
            if (keyB == false && bomb > 0)
            {
                for (int i = 0; i < 4; i++)
                {
                    RemoveStep(0);
                }

                bomb--;

                // 効果音
                audioSource.PlayOneShot(sndBomb);
                UpdateUI();
            }
            keyB = true;
        }
        else
        {
            keyB = false;
        }

        // スペースキーで落下処理(タイマーイベントを呼び出す)
        if (Input.GetKey(KeyCode.Space))
        {
            if (keySpace == false)
            {
                TimeEvent();
            }
            keySpace = true;
        }
        else
        {
            keySpace = false;
        }

        // ESCキーでゲーム終了
        if (Input.GetKey(KeyCode.Escape))
        {
            #if UNITY_EDITOR
                        UnityEditor.EditorApplication.isPlaying = false;
            #elif UNITY_STANDALONE
                  UnityEngine.Application.Quit();
            #endif
        }

    }


        int PosToOrder(int x, int y , int z)
    {
        // フィールド座標からデータ位置に変換する関数
        int ret = 0;

        if(x >= FIELD_X || y >= FIELD_Y || z >= FIELD_Z)
        {
            return - 1;
        }

        ret += x;
        ret += z * FIELD_X;
        ret += y * FIELD_X * FIELD_Z;
        return ret;
    }

    public bool IsMove(int mx, int my, int mz)
    {
        /* ブロックセットが指定座標に移動できるか判定する関数 */
        bool ret = true;
        List<Vector3> list = new List<Vector3>();
        list = bs.GetActiveBlockArray(mx, my, mz);
           
        // 個々のブロックの当たり判定、場外判定を行う
        for(int i=0; i<list.Count; i++)
        {
            if (list[i].x < 0 || list[i].x >= FIELD_X)
                return false;

            if (list[i].y < 0 || list[i].y >= FIELD_Y + 5)
                return false;

            if (list[i].z < 0 || list[i].z >= FIELD_Z)
                return false;

            int order = PosToOrder((int)list[i].x, (int)list[i].y, (int)list[i].z);
            if(order != -1)
            {
                if(list_block[order].activeInHierarchy == true)
                {
                    return false;
                }
            }
        }

        return ret;
    }


    public bool IsDummyMove(int mx, int my, int mz)
    {
        // ダミーを使って移動可能判定をする(BlockSetがダミー変数を持つ)
        bool ret = true;
        List<Vector3> list = new List<Vector3>();
        list = bs.GetDummyBlockArray(mx, my, mz);

        for (int i = 0; i < list.Count; i++)
        {
            if (list[i].x < 0 || list[i].x >= FIELD_X)
                return false;

            if (list[i].y < 0 || list[i].y >= FIELD_Y + 5)
                return false;

            if (list[i].z < 0 || list[i].z >= FIELD_Z)
                return false;

            int order = PosToOrder((int)list[i].x, (int)list[i].y, (int)list[i].z);
            if (order != -1)
            {
                if (list_block[order].activeInHierarchy == true)
                {
                    return false;
                }
            }
        }

        return ret;
    }

    public bool IsFilledStep(int y)
    {
        /* フィールド上の指定の段が埋まったかどうか調べる関数 */
        int x, z, order;
        for(x=0; x<FIELD_X; x++)
        {
            for(z=0; z<FIELD_Z; z++)
            {
                order = PosToOrder(x, y, z);
                if (list_block[order].activeInHierarchy == false)
                {
                    return false;
                }
            }
        }
        return true;
    }

    public bool RemoveStep(int y)
    {
        /* 指定の段を消去する関数 */

        int x, z, order, order2;

        /* // 上の段をコピーすれば消去の必要はない
        for (x = 0; x < FIELD_X; x++)
        {
            for (z = 0; z < FIELD_Z; z++)
            {
                order = PosToOrder(x, y, z);
                list_block[order].SetActive(false);
            }
        }
        */

        // 段を詰める処理
        for(int i=y; i<FIELD_Y-1; i++)
        {
            for (x = 0; x < FIELD_X; x++)
            {
                for (z = 0; z < FIELD_Z; z++)
                {
                    order = PosToOrder(x, i, z);
                    order2 = PosToOrder(x, i + 1, z);
                    list_block[order].SetActive(list_block[order2].activeInHierarchy);

                    // 色合わせ
                    list_block[order].GetComponent<Renderer>().material.color = list_block[order2].GetComponent<Renderer>().material.color;
                }
            }
        }

        // 一番上の段を消去
        for (x = 0; x < FIELD_X; x++)
        {
            for (z = 0; z < FIELD_Z; z++)
            {
                order = PosToOrder(x, FIELD_Y - 1, z);
                list_block[order].SetActive(false);
            }
        }
        return true;
    }
}
